define( [ 'module'
	, './threex.planets',
	, './threex.atmospherematerial'
	, './threex.atmospheredatgui'
	], function(module){
	// set baseUrl for this plugin
	THREEx.Planets.baseURL	= module.uri+'/../';
});